#include "afflib.h"
#include "afflib_i.h"
#include "aimage.h"

#include "imager.h"


void checkline(char *name,char *segname,
	       char *buf,char *copy,AFFILE *af,int number)
{
    char *cc;
    while(buf[0] && isspace(buf[0])) buf++; // advance buf to end of spaces

    if(strncmp(buf,name,strlen(name))!=0) return;
    cc = buf+strlen(name);		// get a pointer
    while(*cc && isspace(*cc)) cc++;	// scan to end of spaces
    char *dd = index(cc,'\n');		// can we find a \n?
    if(dd) *dd = '\000';		// yes; clear it
    if(number){
	af_write_seg(af,segname,atoi(cc),0,0); // store it in the argument...
    }
    else {
	af_write_seg(af,segname,0,cc,strlen(cc)); // store it in the data area
	if(copy) strcpy(copy,cc);		// make a copy
    }
}




/* FreeBSD-specific ident routines */
#ifdef __FreeBSD__
#include <sys/param.h>
#include <sys/mount.h>

int freebsd_get_ata_params(const char *infile,int *controller,int *channel)
{
  int adnum;
  if(sscanf(infile,"/dev/ad%d",&adnum)==1){
      if(controller) *controller = adnum / 2;
      if(channel)    *channel    = adnum % 2;
      return 0;
  }
  return(-1);
}

#define IDENT_DEFINED
/* ident for freebsd */
void ident(const char *infile,AFFILE *af)
{
    int controller = 0;
    int channel    = 0;
  
    /* Check for a SCSI drive.
     * This needs to be cleaned up; right now it assumes SCSI bus 2.
     */
    int da_dev = -1;

    if(sscanf(infile,"/dev/da%d",&da_dev)==1){
	char checkbuf[64];
	char want[64];
	char buf[256];

	/* Now do the devlist to figure out which device this is... */
	scsi_bus = -1;			// for now
	snprintf(want,sizeof(want),",da%d)",da_dev);
	sprintf(checkbuf,"camcontrol devlist");
	FILE *f = popen(checkbuf,"r");
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		if(strstr(buf,want)){	// this is the correct line
		    char *cc = strstr(buf,"at scbus");
		    if(cc && sscanf(cc,"at scbus%d target %d lun %d (pass%d,da%d)",
				    &scsi_bus,&scsi_tid,&scsi_lun,&scsi_pass,&da_dev)==5){
			break;		// found it!
		    }
		}
	    }
	}
	pclose(f);
	if(scsi_bus==-1) err(1,"can't find SCSI device for %s",infile);

	/* Now we need to inquiry */
	sprintf(checkbuf,"camcontrol inquiry %d:%d",scsi_bus,scsi_tid);
	f = popen(checkbuf,"r");
	char capbuf[65536];
	capbuf[0] = 0;
      
	while(!feof(f)){
	    if(fgets(buf,sizeof(buf),f)){
		strcat(capbuf,buf);
		char *cc = index(buf,'>'); // see if it should be terminated
		if(cc) *cc = 0;
		sprintf(checkbuf,"pass%d: <",scsi_pass);
		checkline(checkbuf,AF_DEVICE_MODEL,buf,device_model,af,0);
		sprintf(checkbuf,"pass%d: Serial Number ",scsi_pass);
		checkline(checkbuf,AF_DEVICE_SN,buf,serial_number,af,0);
	    }
	}
	pclose(f);
	af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
	return;
    }
				     
    /* Check for an IDE drive */
    if(strncmp(infile,"/dev/ad",7)==0 &&
       freebsd_get_ata_params(infile,&controller,&channel)==0){
	char capbuf[65536];
	char buf[256];
#if __FreeBSD_version > 600000
	/* Syntax of command was changed in FreeBSD 6.0 */
	sprintf(buf,"atacontrol cap ad%d",controller*2+channel);
#else
	sprintf(buf,"atacontrol cap ata%d %d",controller,channel);
#endif
	puts(buf);
	capbuf[0] = 0;
	FILE *f = popen(buf,"r");
	while(!feof(f)){
	    buf[0] = 0;
	    buf[255] = 0;
	    fgets(buf,sizeof(buf)-1,f);
	    strcat(capbuf,buf);
	    checkline("device model",AF_DEVICE_MODEL,buf,device_model,af,0);
	    checkline("serial number",AF_DEVICE_SN,buf,serial_number,af,0);
	    checkline("firmware revision",AF_DEVICE_FIRMWARE,
		      buf,firmware_revision,af,0);
	    checkline("cylinders",AF_CYLINDERS,buf,0,af,1);
	    checkline("heads",AF_HEADS,buf,0,af,1);
	    checkline("sectors/track",AF_SECTORS_PER_TRACK,buf,0,af,1);
	}
	pclose(f);
	af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
	return;
    }
}


/* FreeBSD ata attach commands */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    sprintf(attach,"atacontrol attach ata%d",dev);
    sprintf(detach,"atacontrol detach ata%d",dev);
    sprintf(master,"/dev/ad%d",dev*2);
    sprintf(slave,"/dev/ad%d",dev*2+1);
}

/* FreeBSD scsi attach commands. Actually does it and returns 0 if successful,
 * as well as the device.
 */
int  scsi_attach(char *fname,int fname_len,int scsidev) // returns 0 if successful & sets fname
{
    char buf[256];
    char looking[64];

    sprintf(buf,"camcontrol rescan %d",scsidev);
    system(buf);

    sprintf(looking,"scbus%d",scsidev);	// what I'm looking for

    /* Now, see if there is a device on this scsi bus */
    FILE *f = popen("camcontrol devlist","r");
    if(!f) err(1,"camcontrol devlist");
    while(!feof(f)){
	memset(buf,0,sizeof(buf));
	if(fgets(buf,sizeof(buf),f)){
	    char *cc = strstr(buf,looking);
	    if(cc){
		int da_num;
		/* Get the target and lun */
		if(sscanf(cc+strlen(looking)+1,"target %d lun %d (pass%d,da%d)",&scsi_tid,&scsi_lun,&scsi_pass,&da_num)!=4){
		    err(1,"could not parse: '%s'",buf);
		}
		snprintf(fname,fname_len,"/dev/da%d",da_num);
		return 0;		// success!
	    }
	}
    }
    pclose(f);
    return -1;				// failure
	
}

#endif





#ifdef linux
#define IDENT_DEFINED
void ident(const char *infile,AFFILE *af)
{
  if(access(infile,R_OK)==0){
      char capbuf[65536];
      char buf[256];
      sprintf(buf,"hdparm -I %s",infile);
      puts(buf);
      capbuf[0] = 0;
      FILE *f = popen(buf,"r");
      while(!feof(f)){
	  buf[0] = 0;
	  buf[255] = 0;
	  fgets(buf,sizeof(buf)-1,f);
	  strcat(capbuf,buf);
	  checkline("Model Number:",AF_DEVICE_MODEL,buf,device_model,af,0);
	  checkline("Serial Number:",AF_DEVICE_SN,buf,serial_number,af,0);
	  checkline("Firmware Revision:",AF_DEVICE_FIRMWARE,
		    buf,firmware_revision,af,0);
	  checkline("cylinders",AF_CYLINDERS,buf,0,af,1);
	  checkline("heads",AF_HEADS,buf,0,af,1);
	  checkline("sectors/track",AF_SECTORS_PER_TRACK,buf,0,af,1);
      }
      pclose(f);
      af_update_seg(af,AF_DEVICE_CAPABILITIES,0,capbuf,strlen(capbuf),1);
  }
}


/*
 * Here are some parameters you may find interesting:
 *       hdparm -R 0x1700 0 0 /dev/hda   (to remove /dev/hda)
 *       ide2: BM-DMA at 0xb800-0xb807, BIOS settings: hde:DMA, hdf:pio
 */
/* Linux ata attach commands. This should work but, in my experience, it doesn't.  */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    system("/sbin/mknod /tmp/hda c  3 0");
    system("/sbin/mknod /tmp/hdb c  3 64");
    system("/sbin/mknod /tmp/hdc c 22 0");
    system("/sbin/mknod /tmp/hdd c 22 64");
    sprintf(attach,"hdparm -R %d",dev);
    sprintf(detach,"hdparm -U %d",dev);
    sprintf(master,"/dev/ad%d",dev*2);
    sprintf(slave,"/dev/ad%d",dev*2+1);
}

/* Linux scsi attach commands */
/* I can't figure out how to do this. */
void make_scsi_attach_commands(char *cmd_attach,char *cmd_detach, int scsidev)
{
    cmd_attach[0] = 0;
    cmd_detach[0] = 0;
}

int  scsi_attach(char *fname,int fname_len,int scsidev) // returns 0 if successful & sets fname
{
    return -1;				// can't do it
}

#endif


/* If we do not have an ident, make it a stub... */
#ifndef IDENT_DEFINED
void ident(const char *infile,AFFILE *af)
{
}


/* We do not have these defined either... */
void make_ata_attach_commands(char *attach,char *detach,char *master,char *slave,int dev)
{
    attach[0] = 0;
    detach[0] = 0;
    master[0] = 0;
    slave[0]  = 0;
}

int  scsi_attach(char *fname,int fname_len,int scsidev) // returns 0 if successful & sets fname
{
    return -1;				// can't do it
}


#endif

